"""
deTEL python.

deTEL python package
"""
import os
import sys

sys.path.append(os.path.dirname(os.path.realpath(__file__)))

__version__ = "0.1.0"
__author__ = 'Cedric Landerer'
__credits__ = 'Agnes Toth-Petroczy Group'
